import json
import functools

from ulib.validators.common import valid_bool

from contextlog import get_logger


# =====
OK = "OK"
WARN = "WARN"
CRIT = "CRIT"


_ATTR_ON_EVENT = "_golem_on_event"
_ATTR_MATCHERS = "_golem_matchers"


# =====
def on_event(method):
    @functools.wraps(method)
    def wrap(previous, current):
        return method((previous and Event(previous)), Event(current))
    setattr(wrap, _ATTR_ON_EVENT, True)
    if hasattr(method, _ATTR_MATCHERS):
        setattr(wrap, _ATTR_MATCHERS, getattr(method, _ATTR_MATCHERS))
    return wrap


def is_event_handler(method):
    return getattr(method, _ATTR_ON_EVENT, False)


def match_event(*matchers):
    assert len(matchers) > 0, "Required minimum one matcher"

    def decorator(method):
        method_matchers = getattr(method, _ATTR_MATCHERS, [])
        for matcher in matchers:
            method_matchers.append(matcher)
        setattr(method, _ATTR_MATCHERS, method_matchers)
        return method

    return decorator


def check_match(method, previous, current):
    previous = (previous and Event(previous))
    current = Event(current)
    for matcher in getattr(method, _ATTR_MATCHERS, []):
        logger = get_logger(
            method="{}.{}".format(method.__module__, method.__name__),
            previous_event=(previous and previous.get_minimal()),
            current_event=current.get_minimal(),
        )
        try:
            if not matcher(previous, current):
                logger.debug("Event is not matched by %s", matcher)
                return False
        except Exception:
            logger.exception("Matching error matcher %s", matcher)
            return False
    return True


# =====
class IncorrectEventError(Exception):
    pass


# =====
class Event:
    # TODO: Make objects for children

    def __init__(self, raw_event):
        self._raw = raw_event.copy()

    def get_raw(self):
        return self._raw

    def get_minimal(self):
        return {
            "host": self.host,  # pylint: disable=no-member
            "service": self.service,  # pylint: disable=no-member
            "status": self.status,  # pylint: disable=no-member
            "description": self.description,  # pylint: disable=no-member
        }


# http://stackoverflow.com/questions/1454984/how-to-define-properties-in-init
for key in (
    "host", "service", "instance",
    "status", "summary", "description",
    "truncated_up_to", "children", "aggregator", "counts",
):
    setattr(Event, key, property(
        lambda self, key=key: self.get_raw().get(key)
    ))


# =====
def convert_status(golem_status):
    # Case-insensitive cruthch:
    #   ok -> OK
    #   crit, critical -> CRIT
    #   warn, warning -> WARN
    status = golem_status.upper()
    status = {"WARNING": WARN, "CRITICAL": CRIT}.get(status, status)
    if status not in (OK, WARN, CRIT):
        raise IncorrectEventError("Invalid status: {}".format(golem_status))
    return status


def convert_golem_event(data):
    # See ya-wiki: sm/obsolete/protocols/JsonDescription
    _check_required_fields(data, ("object", "eventtype"))
    raw_event = {
        "host": data["object"],
        "service": data["eventtype"],
    }
    if valid_bool(data.get("json", False)):
        raw_event.update(json.loads(data.get("info", "{}")))
        _check_required_fields(raw_event, ("status", "description"))
    else:
        raw_event["status"] = convert_status(data.get("status", "critical"))
        raw_event["description"] = data.get("info", "")
    return raw_event


def _check_required_fields(data, fields):
    missing = set(fields).difference(set(data))
    if len(missing) != 0:
        raise IncorrectEventError("Missing required fields: {}".format(", ".join(missing)))



import sys
import argparse
import importlib
import pkgutil
import threading
import socket
import platform
import functools
import logging
import logging.config
import time
import abc

import yaml

import contextlog
from contextlog import get_logger

from ulib import typetools

from .. import tools
from .. import backends

from .. import optconf
from ..optconf.dumper import make_config_dump
from ..optconf.loaders.yaml import load_file as load_yaml_file

from .. import backdoor


# =====
_config = None


def get_config(check_helpers=()):
    if len(check_helpers) > 0:
        for helper in check_helpers:
            if helper not in _config.helpers.configure:
                raise RuntimeError("Helper '{}' is not configured".format(helper))
    return _config


def init(name, description, args=None, raw_config=None):
    global _config
    assert _config is None, "init() has already been called"

    args_parser = argparse.ArgumentParser(prog=name, description=description)
    args_parser.add_argument("-v", "--version", action="version", version=tools.get_version())
    args_parser.add_argument("-c", "--config", dest="config_file_path", default=None, metavar="<file>")
    args_parser.add_argument("-l", "--level", dest="log_level", default=None)
    args_parser.add_argument("-m", "--dump-config", dest="dump_config", action="store_true")
    options = args_parser.parse_args(args)

    # Load configs
    raw_config = (raw_config or {})
    if options.config_file_path is not None:
        raw_config = load_yaml_file(options.config_file_path)
    scheme = _get_config_scheme()
    config = optconf.make_config(raw_config, scheme)

    # Configure logging
    contextlog.patch_logging()
    contextlog.patch_threading()
    logging.setLogRecordFactory(_ClusterLogRecord)
    logging.captureWarnings(True)
    logging_config = raw_config.get("logging")
    if logging_config is None:
        logging_config = yaml.load(pkgutil.get_data(__name__, "configs/logging.yaml"))
    if options.log_level is not None:
        logging_config.setdefault("root", {})
        logging_config["root"]["level"] = _valid_log_level(options.log_level)
    logging.config.dictConfig(logging_config)

    # Update scheme for backend opts
    backend_scheme = backends.get_backend_class(config.core.backend).get_options()
    typetools.merge_dicts(scheme, {"backend": backend_scheme})
    config = optconf.make_config(raw_config, scheme)

    # Update scheme for selected helpers/modules
    for helper_name in config.helpers.configure:
        helper = importlib.import_module(helper_name)
        get_options = getattr(helper, "get_options", None)
        if get_options is None:
            raise RuntimeError("Helper '{}' requires no configuration".format(helper_name))
        typetools.merge_dicts(scheme, {"helpers": get_options()})

    # Provide global configuration for helpers
    _config = optconf.make_config(raw_config, scheme)

    # Print config dump and exit
    if options.dump_config:
        print(make_config_dump(_config, split_by=((), ("helpers",))))
        sys.exit(0)

    return _config


class _ClusterLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fqdn = self._get_cached_fqdn(int(time.time()) // 60)  # Cached value FQDN, updated every minute
        self.node = platform.uname()[1]  # Nodename from uname

    @staticmethod
    @functools.lru_cache(1)
    def _get_cached_fqdn(_):
        return socket.getfqdn()


class Application(metaclass=abc.ABCMeta):
    def __init__(self, app_name, config):
        self._app_name = app_name
        self._app_config = config[app_name]
        self._config = config
        self._stop_event = threading.Event()
        self._respawns = 0

    def stop(self):
        self._stop_event.set()

    def set_app_state(self, backend, app_state):
        app_state = app_state.copy()
        app_state["respawns"] = self._respawns
        backend.system_apps_state.set_state(self._app_name, app_state)

    def get_backend_object(self):
        return backends.get_backend_class(self._config.core.backend)(**self._config.backend)

    ###

    def run(self):
        logger = get_logger(app=self._app_name)  # App-level context
        if self._config.backdoor.enabled:
            backdoor.start(self._config.backdoor.port)
        self._respawns = 0
        while not self._stop_event.is_set():
            if self._app_config.max_fails is not None and self._respawns >= self._app_config.max_fails + 1:
                logger.critical("Reached the respawn maximum, exiting...")
                return -1
            try:
                logger.critical("Ready to work")
                self.process()
            except KeyboardInterrupt:
                logger.critical("Received Ctrl+C, exiting...")
                return 0
            except Exception:
                logger.critical("Error in main loop, respawn...", exc_info=True)
                logger.warning("Sleeping %f seconds...", self._app_config.fail_sleep)
                time.sleep(self._app_config.fail_sleep)
                self._respawns += 1
        self.end()
        return 0

    @abc.abstractmethod
    def process(self):
        raise NotImplementedError

    def end(self):
        pass


# =====
def _valid_log_level(arg):
    try:
        return int(arg)
    except ValueError:
        return logging._nameToLevel[arg.upper()]  # pylint: disable=protected-access


def _get_config_scheme():
    scheme = {
        "core": {
            "backend": optconf.Option(default="zookeeper", help="Backend plugin"),
            "rules_dir": optconf.Option(default="rules", help="Path to rules root"),
        },

        "backdoor": {
            "enabled": optconf.Option(default=False, help="Enable telnet-based backdoor to Python process"),
            "port": optconf.Option(default=2200, help="Backdoor port"),
        },

        "helpers": {
            "configure": optconf.Option(default=[], help="A list of modules that are configured"),
        },

        "api": {
            "backend_connections": optconf.Option(default=1, help="Maximum number of backend connections"),
            "input_limit": optconf.Option(default=5000, help="Limit of the input queue before 503 error"),
            "delete_timeout": optconf.Option(default=15.0, help="Timeout for stop/delete operation"),
            "gunicorn": optconf.Option(default={}, help="Gunicorn options (workers, max_requests, etc.) "
                                                        " exclude entrypoint-specific (like errorlog, accesslog). "
                                                        " See http://docs.gunicorn.org/en/latest/settings.html"),
        },

        "worker": {
            "max_jobs_sleep": optconf.Option(default=1, help="If we have reached the maximum concurrent jobs - "
                                                             "the process goes to sleep (seconds)"),
            "max_jobs": optconf.Option(default=100, help="The maximum number of job processes"),
        },

        "collector": {
            "done_lifetime": optconf.Option(default=60, help="Seconds to wait before deleting completed job"),
        },
    }
    for app in ("worker", "collector"):
        scheme[app].update({
            "max_fails": optconf.Option(default=None, type=int, help="Number of failures after which the program "
                                                                     "terminates"),
            "fail_sleep": optconf.Option(default=5, help="If processing fails, sleep for awhile and restart (seconds)"),
            "empty_sleep": optconf.Option(default=1, help="Interval after which process will sleep when "
                                                          "there are no jobs (seconds)"),
        })
    return scheme



import logging

import flask
import flask_api.app

import gunicorn.app.base
import gunicorn.glogging

from contextlog import get_logger

from .. import backends
from .. import tools
from .. import api
from .. import backdoor

from ..api.rules import RulesResource

from ..api.jobs import JobsResource
from ..api.jobs import JobControlResource

from ..api.system import StateResource
from ..api.system import InfoResource
from ..api.system import ConfigResource

from ..api.golem import GolemResource

from . import init


# =====
class _Api(flask_api.app.FlaskAPI):
    """
        Versioned REST API.
    """

    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
        # Yes, __init__ not from a base class
        flask.Flask.__init__(self, *args, **kwargs)  # pylint: disable=non-parent-init-called
        self.api_settings = flask_api.app.APISettings(self.config)
        self.jinja_env.filters["urlize_quoted_links"] = flask_api.app.urlize_quoted_links
        self.register_blueprint(flask.Blueprint(
            name="flask-api",
            import_name="flask_api.app",
            url_prefix="/flask-api",
            static_folder="static",
        ))
        self.register_blueprint(flask.Blueprint(
            name="powny-api",
            import_name="powny.core.api",
            template_folder="templates",
        ))
        self._resources = {}
        self.add_url_rule("/", "API's list", self._get_apis)

    def add_url_resource(self, version, url_rule, resource):
        def handler(**kwargs):
            return resource.handler(**kwargs)
        handler.__doc__ = resource.docstring
        self.add_url_rule(
            url_rule,
            resource.name,
            handler,
            methods=resource.methods
        )
        self._resources.setdefault(version, [])
        self._resources[version].append(resource)

    def _get_apis(self):
        """ View available API's """
        return {
            version: [
                {"name": resource.name, "url": api.get_url_for(resource)}
                for resource in resources
                if not resource.dynamic
            ]
            for (version, resources) in self._resources.items()
        }


# =====
def make_app(config):
    """
        Use this functions without arguments to create UWSGI app.
    """

    if config.backdoor.enabled:
        backdoor.start(config.backdoor.port)

    pool = backends.Pool(
        size=config.api.backend_connections,
        backend_name=config.core.backend,
        backend_opts=config.backend,
    )

    loader = tools.make_loader(config.core.rules_dir)

    app = _Api(__name__)
    app.add_url_resource("v1", "/v1/rules", RulesResource(
        pool=pool,
        loader=loader,
    ))
    app.add_url_resource("v1", "/v1/jobs", JobsResource(
        pool=pool,
        loader=loader,
        input_limit=config.api.input_limit,
    ))
    app.add_url_resource("v1", "/v1/jobs/<job_id>", JobControlResource(pool, config.api.delete_timeout))
    app.add_url_resource("v1", "/v1/system/state", StateResource(pool))
    app.add_url_resource("v1", "/v1/system/info", InfoResource(pool))
    app.add_url_resource("v1", "/v1/system/config", ConfigResource(config))
    app.add_url_resource("compat", "/api/compat/golem/submit", GolemResource(
        pool=pool,
        loader=loader,
        input_limit=config.api.input_limit,
    ))

    return app


def run(args=None, config=None):
    get_logger(app="api")  # App-level context
    if config is None:
        config = init(__name__, "Powny HTTP API", args)
    _Unicorn(config).run()


# =====
class _Unicorn(gunicorn.app.base.BaseApplication):
    def __init__(self, config):
        self._config = config
        super().__init__()

    def init(self, parser, opts, args):
        pass  # Makes pylint happy

    def load_config(self):
        for (option, value) in self._config.api.gunicorn.items():
            self.cfg.set(option, value)
        self.cfg.set("logger_class", _UnicornLogger)
        self.cfg.set("accesslog", "-")  # For working accesslog (see gunicorn/glogging.py:270)

    def load(self):
        app = make_app(self._config)
        get_logger().critical("Ready to work on %s", self.cfg.bind)
        return app


class _UnicornLogger(gunicorn.glogging.Logger):
    def __init__(self, cfg):  # pylint: disable=super-init-not-called
        self.error_log = logging.getLogger("gunicorn.error")
        self.access_log = logging.getLogger("gunicorn.access")
        self.cfg = cfg

    def setup(self, _):
        pass  # Don't configure logging using the gunicorn options



import collections

from contextlog import get_logger

from ..core import context
from ..core.backends import make_job_id


# =====
def run_in_context(method, kwargs=None, job_id=None, extra=None, fatal=True):
    if callable(method):
        state = context.dump_call(method, (kwargs or {}))
    else:
        assert isinstance(method, bytes)
        state = method

    backend = _Backend()
    thread = context.JobThread(
        backend=backend,
        job_id=(job_id or make_job_id()),
        state=state,
        extra=extra,
    )
    thread.start()
    thread.join()

    if backend.end.exc is not None:
        if fatal:
            raise RuntimeError(backend.end.exc)
        else:
            get_logger().error(backend.end.exc)

    return _Result(job_id, backend.steps, backend.end)


_Step = collections.namedtuple("_Step", ("job_id", "state", "stack"))
_End = collections.namedtuple("_End", ("job_id", "retval", "exc"))
_Result = collections.namedtuple("_Result", ("job_id", "steps", "end"))


class _Backend:
    def __init__(self):
        self.steps = []
        self.end = None

        class _Stub:
            pass
        self.jobs_process = _Stub()
        self.jobs_process.save_job_state = self._save_job_state
        self.jobs_process.done_job = self._done_job

    def _save_job_state(self, job_id, state, stack):
        self.steps.append(_Step(job_id, state, stack))

    def _done_job(self, job_id, retval, exc):
        self.end = _End(job_id, retval, exc)



from flask import request

from ..tools import get_exposed
from ..tools import make_job_state

from ..golem import check_match
from ..golem import convert_golem_event
from ..golem import IncorrectEventError

from . import get_url_for
from . import Resource
from . import ApiError

from .jobs import JobControlResource


# =====
class GolemResource(Resource):
    name = "Push Golem event"
    methods = ("GET", "POST")
    docstring = "Golem-compat handle"

    def __init__(self, pool, loader, input_limit):
        self._pool = pool
        self._loader = loader
        self._input_limit = input_limit

    def process_request(self):
        try:
            current = convert_golem_event(self._get_event_data())
        except IncorrectEventError as err:
            raise ApiError(400, str(err))

        with self._pool.get_backend() as backend:
            if backend.jobs_control.get_input_size() >= self._input_limit:
                raise ApiError(503, "In the queue is more then {} jobs".format(self._input_limit))
            previous = self._get_previous_and_replace(backend, current)
            (head, exposed) = self._get_exposed(backend)
            result = self._run_handlers(backend, previous, current, head, exposed)
            return (result, ("No matching handler" if len(result) == 0 else "Handlers were launched"))

    def _run_handlers(self, backend, previous, current, head, exposed):
        jobs = self._make_jobs_by_matchers(head, previous, current, exposed)
        if len(jobs) == 0:
            return {}
        else:
            return {
                job_id: {"method": job.method_name, "url": self._get_job_url(job_id)}
                for (job_id, job) in zip(backend.jobs_control.add_jobs(head, jobs), jobs)
            }

    def _get_job_url(self, job_id):
        return get_url_for(JobControlResource, job_id=job_id)

    def _get_event_data(self):
        data = dict(request.args)
        if request.method == "POST":
            data.update(dict(request.data or {}))
        for (key, value) in data.items():
            data[key] = (value[-1] if isinstance(value, (list, tuple)) else value)
        return data

    def _get_exposed(self, backend):
        (head, exposed, _, _) = get_exposed(backend, self._loader)
        if exposed is None:
            raise ApiError(503, "No HEAD or exposed methods")
        return (head, exposed)

    def _get_previous_and_replace(self, backend, current):
        old = backend.cas_storage.replace_value(
            path="__check_state/{}/{}".format(current["host"], current["service"]),
            value=current,
            default=None,
        )[0]
        return old.value

    def _make_jobs_by_matchers(self, head, previous, current, exposed):
        return [
            make_job_state(head, name, method, {"previous": previous, "current": current})
            for (name, method) in exposed.get("handlers", {}).items()
            if check_match(method, previous, current)
        ]



import os
import threading
import time

from contextlog import get_logger

from ...core.backends import (
    DeleteTimeoutError,
    JobState,
    make_job_id,
    CasNoValueError,
    CasVersionError,
    CasNoValue,
    CasData,
)
from ...core.tools import (
    make_isotime,
    from_isotime,
    get_node_name,
)

from . import zoo


# =====
_PATH_INPUT_QUEUE = "/input_queue"
_PATH_SYSTEM = "/system"
_PATH_RULES_HEAD = zoo.join(_PATH_SYSTEM, "rules_head")
_PATH_APPS_STATE = zoo.join(_PATH_SYSTEM, "apps_state")
_PATH_JOBS = "/jobs"
_PATH_USER = "/user"
_PATH_CAS_STORAGE = zoo.join(_PATH_USER, "cas_storage")


def _get_path_job(job_id):
    return zoo.join(_PATH_JOBS, job_id)


def _get_path_job_lock(job_id):
    return zoo.join(_get_path_job(job_id), "lock")


def _get_path_job_taken(job_id):
    return zoo.join(_get_path_job(job_id), "taken")


def _get_path_job_state(job_id):
    return zoo.join(_get_path_job(job_id), "state")


def _get_path_job_delete(job_id):
    return zoo.join(_get_path_job(job_id), "delete")


def _get_path_app_state(app_name, node_name):
    return zoo.join(_PATH_APPS_STATE, "{}@{}".format(app_name, node_name))


def _parse_app_state_node(node_name):
    return tuple(node_name.split("@"))


def _get_path_cas_storage(path):
    return zoo.join(_PATH_CAS_STORAGE, path)


def _get_path_cas_storage_lock(path):
    return zoo.join(_get_path_cas_storage(path), "__lock__")


def _make_lock_info(label):
    return {
        "from":     label,
        "when":     make_isotime(),
        "instance": {
            "node": get_node_name(),
            "pid": os.getpid(),
        },
    }


# =====
def init(client):
    logger = get_logger()
    for path in (
        _PATH_INPUT_QUEUE,
        _PATH_SYSTEM,
        _PATH_RULES_HEAD,
        _PATH_APPS_STATE,
        _PATH_JOBS,
        _PATH_USER,
        _PATH_CAS_STORAGE,
    ):
        try:
            with client.make_write_request("backend::init()::create({})".format(path)) as request:
                request.create(path)
            logger.debug("Backend init: path '%s' has been created", path)
        except zoo.NodeExistsError:
            logger.debug("Backend init: path '%s' is already exists", path)


# =====
class JobsControl:
    """
        Interface to managing the jobs (add, delete, get metadata).
        Needs for powny.core.apps.api.
    """

    def __init__(self, client):
        self._client = client
        self._input_queue = self._client.get_queue(_PATH_INPUT_QUEUE)

    def get_jobs_list(self):
        return self._client.get_children(_PATH_JOBS)

    def get_input_size(self):
        return len(self._input_queue)

    def get_jobs_count(self):
        return self._client.get_children_count(_PATH_JOBS)

    def add_jobs(self, head, jobs):
        now = make_isotime()
        added_ids = []
        with self._client.make_write_request("add_jobs()") as request:
            for job in jobs:
                job_id = make_job_id()
                get_logger().info("Registering job", job_id=job_id, head=head,
                                  method=job.method_name, kwargs=job.kwargs)
                request.create(_get_path_job(job_id), {
                    "head": head,
                    "method": job.method_name,
                    "kwargs": job.kwargs,
                    "created": now,
                })
                request.create(_get_path_job_state(job_id), {
                    "state": job.state,
                    "stack": None,
                    "finished": None,
                    "retval": None,
                    "exc": None,
                })
                self._input_queue.put(request, job_id)
                added_ids.append(job_id)
        return added_ids

    def delete_job(self, job_id, timeout=None):
        logger = get_logger(job_id=job_id)
        logger.info("Deleting job")
        try:
            with self._client.make_write_request("delete_job()") as request:
                request.create(_get_path_job_delete(job_id), make_isotime())
        except zoo.NodeExistsError:
            pass  # Lock on existent delete-op
        except zoo.NoNodeError:
            return False
        wait = threading.Event()
        if self._client.exists(_get_path_job_delete(job_id), watch=lambda _: wait.set()):
            wait.wait(timeout=timeout)
            if not wait.is_set():
                msg = "The job was not removed, try again"
                logger.error(msg)
                raise DeleteTimeoutError(msg)
        logger.info("Deleted job")
        return True

    def get_job_info(self, job_id):
        try:
            job_info = self._client.get(_get_path_job(job_id))  # init info

            job_info["deleted"] = self._client.get(_get_path_job_delete(job_id), None)
            job_info["locked"] = self._client.get(_get_path_job_lock(job_id), None)
            job_info["taken"] = self._client.get(_get_path_job_taken(job_id), None)

            state_info = self._client.get(_get_path_job_state(job_id))
            state_info.pop("state", None)  # Remove state
            job_info["finished"] = state_info.pop("finished")
            job_info.update(state_info)  # + stack OR retval OR exc

            return job_info
        except zoo.NoNodeError:
            return None


class JobsProcess:
    """
        Interface to processing the jobs.
        Needs for powny.core.apps.worker.
    """

    def __init__(self, client):
        self._client = client
        self._input_queue = self._client.get_queue(_PATH_INPUT_QUEUE)

    def get_ready_jobs(self):
        for job_id in self._input_queue:
            job_info = self._client.get(_get_path_job(job_id))
            exec_info = self._client.get(_get_path_job_state(job_id))

            with self._client.make_write_request("get_ready_jobs()") as request:
                lock = self._client.get_lock(_get_path_job_lock(job_id))
                lock.acquire(request, _make_lock_info("get_ready_jobs()"))
                request.create(_get_path_job_taken(job_id), make_isotime())
                self._input_queue.consume(request)

            yield JobState(
                head=job_info["head"],
                method_name=job_info["method"],
                kwargs=job_info["kwargs"],
                state=exec_info["state"],
                job_id=job_id,
            )

    def associate_job(self, job_id):
        with self._client.make_write_request("associate_job()") as request:
            # Assign lock to current process
            lock = self._client.get_lock(_get_path_job_lock(job_id))
            lock.release(request)
            lock.acquire(request, _make_lock_info("associate_job()"))

    def release_job(self, job_id):
        with self._client.make_write_request("release_job()") as request:
            self._client.get_lock(_get_path_job_lock(job_id)).release(request)

    def is_deleted_job(self, job_id):
        return self._client.exists(_get_path_job_delete(job_id))

    def save_job_state(self, job_id, state, stack):
        with self._client.make_write_request("save_job_state()") as request:
            request.set(_get_path_job_state(job_id), {
                "state":    state,
                "stack":    stack,
                "finished": None,
                "retval":   None,
                "exc":      None,
            })

    def done_job(self, job_id, retval, exc):
        with self._client.make_write_request("done_job()") as request:
            request.set(_get_path_job_state(job_id), {
                "state":    None,
                "stack":    None,
                "finished": make_isotime(),
                "retval":   retval,
                "exc":      exc,
            })
            self._client.get_lock(_get_path_job_lock(job_id)).release(request)


class JobsGc:
    """
        Interface for garbage collector.
        Needs for powny.core.apps.collector.
    """

    def __init__(self, client):
        self._client = client
        self._input_queue = self._client.get_queue(_PATH_INPUT_QUEUE)

    def get_jobs(self, done_lifetime):
        for job_id in self._client.get_children(_PATH_JOBS):
            to_delete = self._client.exists(_get_path_job_delete(job_id))
            taken = self._client.exists(_get_path_job_taken(job_id))
            lock = self._client.get_lock(_get_path_job_lock(job_id))

            if (to_delete or taken) and not lock.is_locked():
                try:
                    finished = self._client.get(_get_path_job_state(job_id))["finished"]
                except zoo.NoNodeError:
                    continue
                if to_delete or finished is None or from_isotime(finished) + done_lifetime <= time.time():
                    try:
                        with self._client.make_write_request("get_unfinished_jobs()") as request:
                            lock.acquire(request, _make_lock_info("get_unfinished_jobs()"))
                    except (zoo.NoNodeError, zoo.NodeExistsError):
                        continue
                    yield (job_id, to_delete or finished is not None)  # (id, done)

    def push_back_job(self, job_id):
        with self._client.make_write_request("push_back_job()") as request:
            request.delete(_get_path_job_taken(job_id))
            request.delete(_get_path_job_lock(job_id))
            self._input_queue.put(request, job_id)

    def remove_job_data(self, job_id):
        with self._client.make_write_request("remove_job_data()") as request:
            for path_maker in (
                _get_path_job_delete,
                _get_path_job_taken,
            ):
                path = path_maker(job_id)
                if self._client.exists(path):
                    request.delete(path)
            request.delete(_get_path_job_lock(job_id))
            request.delete(_get_path_job_state(job_id))
            request.delete(_get_path_job(job_id))


class Rules:
    """
        Interface to managing the rules HEAD.
    """

    def __init__(self, client):
        self._client = client

    def set_head(self, head):
        with self._client.make_write_request("set_head()") as request:
            request.set(_PATH_RULES_HEAD, head)

    def get_head(self):
        head = self._client.get(_PATH_RULES_HEAD)
        if head is zoo.EmptyValue:
            return None
        else:
            return head


class AppsState:
    """
        Interface to the system statistics about the internal processes
        (like the worker and collector).
    """

    def __init__(self, client):
        self._client = client

    def set_state(self, app_name, app_state):
        state = {
            "when": make_isotime(),
            "pid": os.getpid(),
            "state": app_state,
        }
        self._set_raw_state(app_name, get_node_name(), state)

    def _set_raw_state(self, app_name, node_name, state):
        path = _get_path_app_state(app_name, node_name)
        try:
            with self._client.make_write_request("set_state()") as request:
                request.set(path, state)
        except zoo.NoNodeError:
            with self._client.make_write_request("create_state_node()") as request:
                request.create(path, state, ephemeral=True)

    def get_full_state(self):
        full_state = {}
        for app_state_node in self._client.get_children(_PATH_APPS_STATE):
            (app_name, node_name) = _parse_app_state_node(app_state_node)
            try:
                state = self._client.get(_get_path_app_state(app_name, node_name))
            except zoo.NoNodeError:
                continue
            full_state.setdefault(app_name, {})
            full_state[app_name][node_name] = state
        return full_state


class CasStorage:
    """
        Interface to abstract CAS storage for user scripts.
    """

    def __init__(self, client):
        self._client = client

    def set_value(self, path, value, version=None):
        try:
            self.replace_value(path, value=value, version=version, default=None)
            return True
        except CasVersionError:
            get_logger().exception("Can't set '%s' value with version %s", path, version)
            return False

    def get_value(self, path, default=CasNoValue):
        return self.replace_value(path, value=CasNoValue, default=default)[0]

    def replace_value(self, path, value=CasNoValue, version=None, default=CasNoValue, fatal_write=True):
        """
            replace_value() - implementation of the CAS, stores the new value if it is superior to the
                              existing version. Standard kazoo set() require strict comparison and
                              incremented version of the data themselves.

            If:
                value == CasNoValue -- read operation
                value == ... -- write the new value and return the old

                version is None -- write
                version is not None -- write if version >= old_version
        """

        path = _get_path_cas_storage(path)

        try:
            with self._client.make_write_request("cas_ensure_path()") as request:
                request.create(path, recursive=True)
        except zoo.NodeExistsError:
            pass

        with self._client.get_lock(_get_path_cas_storage_lock(path)):
            old = self._client.get(path)
            if old is zoo.EmptyValue:
                if default is CasNoValue:
                    raise CasNoValueError()
                old = CasData(value=default, version=None, stored=None)
            else:
                old = CasData(
                    value=old["value"],
                    version=old["version"],
                    stored=from_isotime(old["stored"]),
                )

            if value is not CasNoValue:
                if version is not None and old.version is not None and version <= old.version:
                    write_ok = False
                    msg = "Can't rewrite '{}' with version {} (old version: {})".format(path, version, old.version)
                    if fatal_write:
                        raise CasVersionError(msg)
                    else:
                        get_logger().debug(msg)
                else:
                    with self._client.make_write_request("cas_save()") as request:
                        request.set(path, {
                            "value":   value,
                            "version": version,
                            "stored":  make_isotime(),
                        })
                        write_ok = True
            else:
                write_ok = None

        return (old, write_ok)



import contextlib
import textwrap

from ..core import apps
from .tmpfile import write_file


# =====
@contextlib.contextmanager
def configured(text=None):
    if text is not None:
        with write_file(textwrap.dedent(text)) as file_path:
            config = apps.init("test", "Test case", ["-c", file_path])
    else:
        config = apps.init("test", "Test case", [])
    try:
        yield config
    finally:
        apps._config = None  # pylint: disable=protected-access



import tabloid
import colorama
import pygments
import pygments.lexers
import pygments.formatters

from . import Section


# =====
def make_config_dump(config, split_by, width=None):
    table = tabloid.FormattedTable(width=width, header_background=colorama.Back.BLUE)
    table.add_column("Option", _format_option)
    table.add_column("Value", _format_value)
    table.add_column("Default", _format_default_value)
    table.add_column("Help")

    for row in _make_plain_dump(config, tuple(map(tuple, split_by))):
        if row is None:
            table.add_row((" ") * 4)
        else:
            table.add_row(row)
    return "\n".join(table.get_table())


def _format_option(name, row):
    if row[1] != row[2]:
        return "{}{}{}".format(colorama.Fore.RED, name, colorama.Style.RESET_ALL)
    return name


def _format_value(value, _):
    return pygments.highlight(
        value,
        pygments.lexers.PythonLexer(),
        pygments.formatters.TerminalFormatter(bg="dark"),
    ).replace("\n", "")


def _format_default_value(value, row):
    if row[1] != row[2]:
        return _format_value(value, None)
    else:
        return " " * len(value)


def _make_plain_dump(config, split_by=(), path=()):
    plain = []
    for (key, value) in config.items():
        if isinstance(value, Section):
            if len(plain) != 0 and path in split_by:
                plain.append(None)
            plain += _make_plain_dump(value, split_by, path + (key,))
        else:
            default = config._get_default(key)  # pylint: disable=protected-access
            plain.append((
                ".".join(path + (key,)),
                repr(value),
                repr(default),
                config._get_help(key),  # pylint: disable=protected-access
            ))
    return plain



import platform
import datetime
import calendar
import time

import dateutil.parser
import pkginfo

from contextlog import get_logger

from . import context
from . import imprules
from . import golem

from .backends import JobState


# =====
def get_node_name():
    return platform.uname()[1]


def get_version():
    pkg = pkginfo.get_metadata("powny")
    return (pkg.version if pkg is not None else "<unknown>")


def get_user_agent():
    return "Powny/{}".format(get_version())


# =====
def make_isotime(unix=None):  # ISO 8601
    if unix is None:
        unix = time.time()
    return datetime.datetime.utcfromtimestamp(unix).strftime("%Y-%m-%d %H:%M:%S.%fZ")


def from_isotime(line):
    dt = dateutil.parser.parse(line)
    return calendar.timegm(dt.utctimetuple()) + dt.microsecond / 10 ** 6  # pylint: disable=maybe-no-member


# =====
def make_loader(rules_root):
    return imprules.Loader(
        prefix=rules_root,
        group_by=(
            ("handlers", golem.is_event_handler),
            ("methods", lambda _: True),
        ),
    )


def get_exposed(backend, loader):
    head = backend.rules.get_head()
    exposed = None
    errors = None
    exc = None
    if head is not None:
        try:
            (exposed, errors) = loader.get_exposed(head)
        except Exception as err:
            exc = "{}: {}".format(type(err).__name__, err)
            get_logger().exception("Can't load HEAD '%s'", head)
    return (head, exposed, errors, exc)


def make_job_state(head, name, method, kwargs):
    return JobState(
        head=head,
        method_name=name,
        kwargs=kwargs,
        state=context.dump_call(method, kwargs),
        job_id=None,
    )
